from pyviko import core, restriction

class OverGene:
	'''
	Class for the overprinted gene.
	'''
	frame = 1
	startNucleotideIndex = -1
	preSequence = '' # includes 1-2nt removed by core.findOverGene
	geneSequence = ''
	postSequence = '' # includes 1-2nt removed by core.findOverGene
	overAAs = ''
	
	def __init__(self, overSeq, startNtIndex, seq, frameOver = 1):
		if overSeq != '':
			ol = core.findOverlap(seq, overSeq)
			if ol[0] == 0: #overprinted gene starts before
				startNtIndex = -1
				frameOver = 4-(ol[1]%3)
				self.preSequence = overSeq[ol[1]-(-frameOver+4):ol[1]] + seq[:3-(-frameOver+4)]
			else: #overprinted gene starts after
				startNtIndex = ol[0]
		self.geneSequence = overSeq
		self.startNucleotideIndex = startNtIndex
		self.combSequence = seq
		self.frame = frameOver
		self.overAAs = core.translate(self.preSequence + core.seqify(core.findOverprintedGene(seq, startNtIndex, frameOver)) + self.postSequence)

class Mutant:
	'''
	Class for target gene mutagenesis.
	'''
	seq = ''
	codons = []
	nMut = 1
	overGene = False
	mutants = []
	
	def __init__(self, sequence, numMutations = 1, regEx = False):
		self.seq = sequence
		self.nMut = numMutations
		self.codons = core.codonify(sequence)
		self.regex = regEx
		
	def setOverGene(self, overSeq = '', startNtIndex = -1, overFrame = 1):
		'''
		Adds the overprinted gene to the current `Mutant` object.
		'''
		if overSeq == '' and startNtIndex == -1 and overFrame == 1:
			raise core.SequenceError("You must provide either the sequence of an overprinted gene, or its start position/frame in the knockout sequence.")
		else:
			self.overGene = OverGene(overSeq, startNtIndex, self.seq, overFrame)
		
	def vector(self, sequence):
		'''
		Adds the vector sequence to the current `Mutant` object 
		(primarily for making primers of early knockouts).
		'''
		if self.seq in sequence:
			self.vectorSeq = sequence
		else:
			raise core.SequenceError("Could not find target sequence in vector sequence.")
			
	def findMutants(self, ignoreRxSites = True, rSiteLength = 6, rSites = restriction.defaultEnzymes()):
		'''
		Returns a list of mutants that add a premature stop codon 
		(or mutate the start codon) without changing the overprinted 
		gene, and which add or remove a restriction site.
		'''
		stops = findPossibleStopCodons(self.codons, self.nMut)

		if self.overGene != False:
			if len(self.overGene.geneSequence) > 0:
				stops = mutateStartCodon(self.codons, self.nMut) + stops
			safeMutations = []
			newPreSequence = '';
			for poss in stops:
				nCodons = [codon for codon in self.codons]
				newCodons = core.insertMutation(nCodons, poss)
				if self.overGene.geneSequence != '':
					newPreSequence = self.overGene.preSequence[:4-self.overGene.frame] + newCodons[0][:self.overGene.frame - 1]
				newOverAAs = core.translate(newPreSequence + core.seqify(core.findOverprintedGene(core.seqify(newCodons), self.overGene.startNucleotideIndex, self.overGene.frame)))
				if newOverAAs == self.overGene.overAAs:
					safeMutations.append(poss)
		else:
			safeMutations = stops
		finalWinners = [s for s in safeMutations]
		if not ignoreRxSites:
			### Two approaches: regex and non-regex.
			restrictionSiteLengths = list(set([len(k) for k in rSites.keys()]))
			
			if rSiteLength == 'all':
				tempRestrictionSites = rSites
			elif rSiteLength >= min(restrictionSiteLengths) and rSiteLength <= max(restrictionSiteLengths):
				rKeys = [k for k in rSites.keys() if len(k) == rSiteLength]
				tempRestrictionSites = {} # Reduce size of dict. searched
				for site in rKeys:
					tempRestrictionSites[site] = rSites[site]
				restrictionSiteLengths = [rSiteLength]
			else:
				raise core.SequenceError("Invalid restriction site length.")		
			
			newSites = [] # list of lists
			
			### Regex:
			if self.regex:
				baseSites = restriction.reFindEnzymes(self.seq)
				
				for mut in safeMutations:
					newSites.append([])
					newSites[-1] += restriction.reFindEnzymes(core.seqify(core.insertMutation(self.codons, mut)))
					
			### Non-regex:
			else:
				
				baseSites = []
				for length in restrictionSiteLengths:
					baseSites += restriction.findNcutters(self.seq, length)
					
				for mut in safeMutations:
					newSites.append([])
					for length in restrictionSiteLengths:
						newSites[-1] += restriction.findNcutters(core.seqify(core.insertMutation(self.codons, mut)), length)
			
			winners = {}
			for l in newSites:
				if l != baseSites: #this is why I should use sets
					tempSites = [c for c in baseSites]
					tempAddedSites = []
					for site in l: #basically, removing everything in the new list from the old list to get the differences
						try:
							tempSites.remove(site)
						except ValueError:
							tempAddedSites.append(site)
					
					for i in range(0,len(tempSites)):
						tempSites[i] = (tempSites[i][0], tempSites[i][1], '-')
						
					for i in range (0, len(tempAddedSites)):
						tempAddedSites[i] = (tempAddedSites[i][0], tempAddedSites[i][1], '+')
					
					winners[safeMutations[newSites.index(l)]] = tempSites + tempAddedSites
			
			finalWinners = [(x,winners[x]) for x in sorted(winners.keys(), key=lambda x: x[0])]		
			
		return finalWinners
	
def findPossibleStopCodons(codons, n):
	'''
	Given a list `codons`, finds individual codons that
	can be mutated to a stop codon given `n` mutations. Returns 
	a list of tuples of the form `(index, 'codon')` where `'codon'`
	is the mutated codon. 
	'''
	#TODO: can optimize this as in js
	if codons[-1] in core.stopCodons: 
		codons = codons[:-1] #remove c-terminal stop codon
	
	almostStopCodons = {}
	#build dict of codons that can be mutated to a stop codon 
	
	for c in core.stopCodons:
		for i in list(range(0,3)):
			for nt in 'ACTG':
				if c[:i]+nt+c[i+1:] not in core.stopCodons:
					try:
						almostStopCodons[c[:i]+nt+c[i+1:]].append(c)
					except KeyError:
						almostStopCodons[c[:i]+nt+c[i+1:]] = [c]
	
	if n == 2:
		for c in almostStopCodons.keys():
			for i in list(range(0,3)):
				for nt in 'ACTG':
					if c[:i]+nt+c[i+1:] not in core.stopCodons:
						try:
							almostStopCodons[c[:i]+nt+c[i+1:]] += almostStopCodons[c]
						except KeyError:
							 almostStopCodons[c[:i]+nt+c[i+1:]] = almostStopCodons[c]
		for i in almostStopCodons.keys():
			almostStopCodons[i] = list(set(almostStopCodons[i]))
	
	# creates a list of tuples of the form (index, ['list', 'of', 'stop', 'codons']) to be further pre-processed	
	preMatches = [(i, almostStopCodons[codons[i]]) for i in range(0,len(codons)) if codons[i] in almostStopCodons.keys()]
	
	matches = []  
	
	# further processing to create actual tuples (index, 'codon')	
	for m in preMatches:
		for codon in m[1]:
			matches.append((m[0],codon))
			
	return matches

def mutateStartCodon(codons, n):
	'''
	Given a list `codons`, makes up to `n` mutations (n<2)
	to destroy the start codon. Returns a formatted list
	of tuples in the form `(index, 'mutated codon')`.
	'''
	start = codons[0]
	muts = []
	for i in list(range(0,3)):
		for nt in 'ACTG':
			mutCodon = start[:i]+nt+start[i+1:]
			if mutCodon != start and mutCodon != 'ATG':
				muts.append(mutCodon)
	newMuts = [z for z in muts]
	if n == 2:
		for e in muts:
			for i in list(range(0,3)):
				for nt in 'ACTG':
					mutCodon = e[:i]+nt+e[i+1:]
					if mutCodon != start and mutCodon != 'ATG' and mutCodon not in newMuts:
						newMuts.append(mutCodon)
					
	return [(0,m) for m in newMuts]



stopCodons = ['TAG', 'TAA', 'TGA']
translation = {'CTT': 'L', 'ATG': 'M', 'AAG': 'K', 'AAA': 'K', 'ATC': 'I', 'AAC': 'N', 'ATA': 'I', 'AGG': 'R', 'CCT': 'P', 'ACT': 'T', 'AGC': 'S', 'ACA': 'T', 'AGA': 'R', 'CAT': 'H', 'AAT': 'N', 'ATT': 'I', 'CTG': 'L', 'CTA': 'L', 'CTC': 'L', 'CAC': 'H', 'ACG': 'T', 'CAA': 'Q', 'AGT': 'S', 'CAG': 'Q', 'CCG': 'P', 'CCC': 'P', 'TAT': 'Y', 'GGT': 'G', 'TGT': 'C', 'CGA': 'R', 'CCA': 'P', 'TCT': 'S', 'GAT': 'D', 'CGG': 'R', 'TTT': 'F', 'TGC': 'C', 'GGG': 'G', 'GGA': 'G', 'TGG': 'W', 'GGC': 'G', 'TAC': 'Y', 'GAG': 'E', 'TCG': 'S', 'TTA': 'L', 'GAC': 'D', 'TCC': 'S', 'GAA': 'E', 'TCA': 'S', 'GCA': 'A', 'GTA': 'V', 'GCC': 'A', 'GTC': 'V', 'GCG': 'A', 'GTG': 'V', 'TTC': 'F', 'GTT': 'V', 'GCT': 'A', 'ACC': 'T', 'TTG': 'L', 'CGT': 'R', 'CGC': 'R'}

import warnings, os
#from future import print_function

class SequenceError(Exception):
	pass

def codonify(sequence):
	'''
	Converts an input DNA sequence (str) to a list of codons.
	'''
	if type(sequence) == type([]):
		return sequence
	return [sequence[i:i+3] for i in list(range(0,len(sequence),3))]

def seqify(cod):
	'''
	Converts an input list of codons into a DNA sequence (str).
	'''
	if type(cod) == type("str"):
		return cod
	sequence = ""
	for codon in cod:
		sequence += codon
	return sequence
	
def translate(codons):
	'''
	Translates a list of DNA codons into the corresponding amino 
	acids, stopping translation if a stop codon is encountered.
	'''
	codons = codonify(codons)
	aa = ''
	for i in list(range(0,len(codons))):
		if codons[i] in stopCodons or len(codons[i]) != 3:
			#codons = codons[:i]
			if codons[i] in stopCodons:
				aa = aa + '*'
			break
		try:
			aa = aa + translation[codons[i]]
		except KeyError as e:
			raise SequenceError("Invalid codon: " + e.message)
	return aa

def insertMutation(codons, mut):
	'''
	Takes as input a list of
	`codons` to mutate and a tuple `mut` in the form 
	(index, 'mutated codon') ex. `(3, 'TAA')`.
	Returns a list of codons with a mutation generated by 
	removing the codon to be mutated from the list, then adding
	the newly mutated codon in its position. 
	'''
	iCodons = [c for c in codons]
	iCodons[mut[0]] = mut[1]	
	return iCodons
	
def pointMutant(seq, mut):
	'''
	Takes as input a sequence `seq` to mutate
	and a tuple `mut` in the form (index, 'mutated nt') ex. `(3, 'A')`.
	Returns a nucleotide sequence with a point mutation.
	'''
	return seq[:mut[0]] + mut[1] + seq[mut[0]+1:]
	
def findOverprintedGene(seq, startIndex, frame=1):
	'''
	Given a sequence `seq` and the `startIndex` of 
	an overprinted gene, returns a list of codons that
	correspond to the overprinted gene. The `frame` 
	argument is only necessary if the overprinted
	gene's start codon is before the input sequence, in 
	which case `startIndex` must be -1. <br> <br>
	**NOTE:** the index of the first nucleotide in `seq` is 0,
	so if the overprinted gene starts from the 59th nucleotide
	of `seq`, the `startIndex` will be 58.
	'''
	
	if startIndex != -1:
		frame = 1   # In case `frame` argument provided erroneously
		codons = codonify(seq[startIndex:])[:-1] # Remove last (incomplete) codon
	else:
		if frame == 1:
			raise SequenceError("The overprinted sequence is in the same frame as the main coding sequence. Please provide a frame argument.")
		codons = codonify(seq[frame - 1:])[:-1] # Remove last (incomplete) codon
	for i in list(range(1,len(codons))):
		if codons[i] in stopCodons:
			codons = codons[:i]
			break
		
	if codons[0] != 'ATG' and startIndex != -1:
		#NOTE: Not all viral genes are initiated with ATG.
		warnings.warn("The first codon of your sequence is not a start codon.")
		
	return codons

def reverseComplement(seq):
	'''
	Given a sequence `seq`, returns the reverse complement.
	'''
	seq = seqify(seq)
	pairs = {'A':'T', 'T':'A', 'C':'G', 'G':'C'}
	rev = ""
	#Here should add reverse complements for regex sites? i.e. Y -> R 
	try:	
		for nt in seq[::-1]:
			rev += pairs[nt]
	except KeyError:
		print("Unknown nucleotide '" + nt  + "' encountered.")
		return "False"
	
	return rev
	
def findOverlap(seq1, seq2):
	'''
	Given two sequences, returns a tuple `(i1, i2)` where `i1`
	is the index in `seq1` where the overlap with `seq2` begins and
	`i2` is the corresponding index in `seq2`.
	'''
	i1 = 0
	i2 = 0
	if seq1 in seq2:
		i2 = seq2.index(seq1)
	elif seq2 in seq1:
		i1 = seq1.index(seq2)
	else:
		max12 = 0
		max21 = 0
		l1 = len(seq1)
		l2 = len(seq2)
		overall = min(l1,l2) + 1
		for i in list(range(1, overall)):
			if seq1[:i] == seq2[l2-i:]:
				max21 = i
			if seq2[:i] == seq1[l1-i:]:
				max12 = i				
		if max12 > max21:
			i1 = l1 - max12
		elif max12 <= max21: #without this small overlaps equal on both sides are mishandled
			i2 = l2 - max21
		else: 
			raise SequenceError("No overlap detected between input sequences")
			
	return (i1, i2)
	
def readFasta(loc):
	'''
	Reads in a FASTA file, returns tuples in the form
	`('> identifying information', 'sequence')`.
	'''
	f = open(loc, 'r')
	seqs = []
	iden = ''
	seq = ''
	for line in f.readlines():
		if iden == '':
			try:
				if line.lstrip()[0] != '>':
					raise SequenceError("Invalid file format: id line doesn't begin with '>'")
				iden = line.strip()
			except IndexError: #blank line
				next
		else:
			try:
				if line.lstrip()[0] == '>':
					seqs.append((iden, seq))
					iden = line.strip()
					seq = ''
				else:
					seq += line.strip().upper().replace(' ', '')
			except IndexError: #blank line
				next
	seqs.append((iden, seq))
	f.close()
	return seqs
	
def writeFasta(fname, mutlist, seq, hasRxSites = False, rloc = "", floc = ""):
	'''
	Given a filename `fname`, list of mutations `mutlist` input sequence `seq`
	and an optional file location  `rloc` (relative location) or `floc` (absolute location), 
	generates a FASTA file with all mutants in the sequence.
	Accepts input in two formats: if hasRxSites=False, assumes the mutations are of the form
	`(mutant codon index, 'stop codon')`.
	'''
	fname = fname.replace('|', '.')[:30]
	for character in fname:
		if character not in '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.,\'"()-_':
			fname = fname.replace(character, ' ')
	fname = ' '.join([i for i in fname.split(' ') if i != '']) + '.fasta'
	if floc == "":
		dirs = os.listdir(os.getcwd())
	else:
		dirs = os.listdir(floc)
	if fname in dirs:
		i = 1
		while fname[:-6] + '(' + str(i) + ').fasta' in dirs:
			i += 1
		fname = fname[:-6] + '(' + str(i) + ').fasta'
	del i
	fasta = open(fname, 'w')
	for m in mutlist:
		fasta.write(">mutant at codon " + str(m[0][0]+1) +': \n' )
		codons = codonify(seq)
		mutSeq = seqify(codons[:m[0][0]]+[m[0][1]]+codons[m[0][0]+1:])
		fasta.write('\n'.join([mutSeq[i:i+100] for i in range(0,len(mutSeq),100)]) +'\n')
		
	fasta.close()
	return True



import setuptools
from distutils.core import setup


#with open("README.md", "r") as fh:
#    long_description = fh.read()

setup(name='pyviko',
	version='1.1.0.6',
	description='Design knockout viruses in Python',
	author='LJ Taylor',
	author_email='l'+'taylor'+str(3+4)+'@'+'tu'+'lane.edu',
	url='https://github.com/louiejtaylor/pyViKO',
	packages=['pyviko'],
	#install_requires=['future'],
	license='MIT License',
	classifiers = ['Intended Audience :: Science/Research',
				'Environment :: Console',
				'Environment :: Web Environment',
				'License :: OSI Approved :: MIT License',
				'Programming Language :: Python :: 3',
				'Topic :: Scientific/Engineering :: Bio-Informatics',],
	py_modules = ['pyviko.core', 'pyviko.mutation', 'pyviko.restriction']
     )
